/*
 *      CUDA.cu - ,   GPU
 *
 *      Copyright 2011 Edward V. Emelianoff <eddy@sao.ru>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */
#define _CUDA_CU_
#include "CUDA.h"
#include "CUtools.h"

#if __CUDA_ARCH__ >= 200 //  >= 2.x
	#define SHMEMSZ	49151
	#define QBLKSZ	32		// QBLKSZ = sqrt(LBLKSZ)
	#define LBLKSZ	1024
#else
	#define SHMEMSZ	16383
	#define QBLKSZ	16		// QBLKSZ = sqrt(LBLKSZ)
	#define LBLKSZ	512
#endif

static float *Sin_d = NULL, *Cos_d = NULL;
static int sincosize = 0;

cudaError_t CUerr;
inline int CUERROR(char *str){
	if(CUerr != cudaSuccess){
		fprintf(stderr, "%s\n", str);
		return 1;
	}else return 0;
}
// ,    (  - )
#define RETMACRO return
//     
#define CUALLOC(var, size)		do{				\
	CUerr = cudaMalloc((void**)&var, size);		\
	if(CUERROR("CUDA: can't allocate memory")){	\
		RETMACRO;								\
}}while(0)
#define CUMOV2DEV(dest, src, size) do{			\
	CUerr = cudaMemcpy(dest, src, size,			\
				cudaMemcpyHostToDevice);		\
	if(CUERROR("CUDA: can't copy data to device")){\
		RETMACRO;}								\
}while(0)
#define CUMOV2HOST(dest, src, size) do{			\
	CUerr = cudaMemcpy(dest, src, size,			\
				cudaMemcpyDeviceToHost);		\
	if(CUERROR("CUDA: can't copy data to host")){\
		RETMACRO;}								\
}while(0)
#define CUFREE(var) cudaFree(var)

//   arr
__global__ void normalize_vec(float *arr, int arrsize){
	__shared__ float max[LBLKSZ];
	int idx = threadIdx.x;
	int blksize = (arrsize + blockDim.x - 1) / blockDim.x;
	int b_beg = idx * blksize;
	if(b_beg >= arrsize) return;
	int b_end = b_beg + blksize;
	if(b_end > arrsize) b_end = arrsize;
	int i; float *ptr = &arr[b_beg];
	float mm = *ptr++;
	for(i = b_beg +1 ; i < b_end; i++, ptr++)
		if(mm < *ptr) mm = *ptr;
	max[idx] = mm;
	__syncthreads();
	if(idx == 0){
		mm = max[0];
		for(i = 1; i < LBLKSZ; i++)
			if(mm < max[i]) mm = max[i];
		max[0] = mm;
	}
	__syncthreads();
	ptr = &arr[b_beg];
	mm = max[0];
	for(i = b_beg ; i < b_end; i++, ptr++)
		*ptr /= mm;
}

__global__ void fill_zeros(float *arr, int arrsize, int W, int H){
	int x = blockIdx.x * blockDim.x + threadIdx.x;
	int y = blockIdx.y * blockDim.y + threadIdx.y;
	if(x > W || y > H) return;
	arr[x + y*W] = 0.f;
}

//    sincos
__global__ void fill_sincos(int angles,
							float *Sin_d, float *Cos_d,
							float anglestep, float conv){
	int k = blockIdx.x * blockDim.x + threadIdx.x;
	if(k >= angles) return;
	float theta = ((float)k / anglestep - 90.f) * conv;
	sincosf(theta, &Sin_d[k], &Cos_d[k]);
}

//      
extern "C" void init_sincos(int angles){
	float anglestep = (float)angles / 270.f; //   
	float conv = M_PI / 180.f;
	int blks = (angles + QBLKSZ - 1) / QBLKSZ;
	int threads = LBLKSZ;
	if(!Sin_d || !Cos_d || angles != sincosize){ //     
		CUFREE(Cos_d);
		CUFREE(Sin_d);
printf("aloca %d\n", angles);
		CUALLOC(Sin_d, angles*sizeof(float));
		CUALLOC(Cos_d, angles*sizeof(float));
		fill_sincos<<<blks, threads>>>(angles, Sin_d, Cos_d, anglestep, conv);
		sincosize = angles;
	}
printf("ready\n");
	cudaThreadSynchronize();
}

/*
 *      
 */
__global__ void fill_lin_hough_array(float *ima_d,
						int imW, int imH,
						float *Sin_d, float *Cos_d,
						int Rmax, int angles, float treshold,
						float *hough_d){
	int xi = blockIdx.x * blockDim.x + threadIdx.x;
	int yi = blockIdx.y * blockDim.y + threadIdx.y;
	int i = xi + imW * yi;
	float x = (float)xi;
	float y = (float)yi;
	int Y = 0, k, R;
	if(xi >= imW || yi >= imH) return;
	float ima = ima_d[i];
	//float tresh = *treshold;
	if(ima > treshold){
		for(k = 0; k < angles; k++, Y+=Rmax){
			// R = x*cos(theta) + y*sin(theta)
			R = (int)(0.5f + x * Cos_d[k] + y * Sin_d[k]);
			if(R > 0 && R < Rmax) hough_d[R + Y] += ima;
		}
	}
}
/* 182
 *      
 * :
 * 		ima -  
 * 		imW, imH -    
 * 		Rmax -   R
 * 		angles -    (    180/angles )
 * : hough -    ,
 * 		     
 * 		!!!         
 *      
 */
extern "C" int fill_hough_lines(float *ima, int imW, int imH, int Rmax, int angles, float *hough){
	#undef RETMACRO
	#define RETMACRO do{ ret = 0; goto free_all; }while(0)
	int sz, ret = 1;
	int lblksz = LBLKSZ;
	float *ima_d = NULL, *hough_d = NULL;
	float treshold = 0.5f;
	init_sincos(angles);
	sz = imW * imH;
	dim3 blkdim(QBLKSZ, QBLKSZ);
	dim3 griddim((imW+QBLKSZ-1)/QBLKSZ, (imH+QBLKSZ-1)/QBLKSZ);
	dim3 hgriddim((Rmax+QBLKSZ-1)/QBLKSZ, (angles+QBLKSZ-1)/QBLKSZ);
	CUALLOC(ima_d, sz*sizeof(float));
	CUMOV2DEV(ima_d, ima, sz*sizeof(float));
	sz = Rmax * angles;
	CUALLOC(hough_d, sz*sizeof(float));
	fill_zeros<<<hgriddim, blkdim>>>(hough_d, sz, Rmax, angles); // 
	cudaThreadSynchronize();
	//~ CUMOV2DEV(hough_d, hough, sz*sizeof(float));
	fill_lin_hough_array<<<griddim, blkdim>>>(ima_d, imW, imH, Sin_d, Cos_d,
									Rmax, angles, treshold, hough_d);
	cudaThreadSynchronize();
	normalize_vec<<<1, lblksz>>>(hough_d, sz);
	cudaThreadSynchronize();
	CUMOV2HOST(hough, hough_d, sz*sizeof(float));
free_all:
	CUFREE(hough_d);
	CUFREE(ima_d);
	return ret;
	#undef RETMACRO
	#define RETMACRO return
}
