/*
 *      CUDA.cu - ,   GPU
 *
 *      Copyright 2011 Edward V. Emelianoff <eddy@sao.ru>
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *      MA 02110-1301, USA.
 */
#define _CUDA_CU_
#include "CUDA.h"
#include "CUtools.h"


#include <sys/time.h>
extern "C" float ftime(){
	float t, add = 0.f;
	struct timeval tv;
	static time_t sec = 0;
	gettimeofday(&tv, NULL);
	if(sec == 0) sec = tv.tv_sec;
	else{
		if(sec != tv.tv_sec){
			add = (tv.tv_sec - sec) * 1e6;
			sec = tv.tv_sec;
		}
	}
	t = add + ((float)tv.tv_usec);
	return t;
}

static __device__ float *Sin_d = NULL, *Cos_d = NULL;
static __device__ int sincosize = 0;

#if __CUDA_ARCH__ >= 200 //  >= 2.x
	#define SHMEMSZ	49151
	#define BLKSZ	1024
#else
	#define SHMEMSZ	16383
	#define BLKSZ	512
#endif
cudaError_t CUerr;
inline int CUERROR(char *str){
	if(CUerr != cudaSuccess){
		fprintf(stderr, "%s\n", str);
		return 1;
	}else return 0;
}
// ,    (  - )
#define RETMACRO return
//     
#define CUALLOC(var, size)		do{				\
	CUerr = cudaMalloc((void**)&var, size);		\
	if(CUERROR("CUDA: can't allocate memory")){	\
		RETMACRO;								\
}}while(0)
#define CUMOV2DEV(dest, src, size) do{			\
	CUerr = cudaMemcpy(dest, src, size,			\
				cudaMemcpyHostToDevice);		\
	if(CUERROR("CUDA: can't copy data to device")){\
		RETMACRO;}								\
}while(0)
#define CUMOV2HOST(dest, src, size) do{			\
	CUerr = cudaMemcpy(dest, src, size,			\
				cudaMemcpyDeviceToHost);		\
	if(CUERROR("CUDA: can't copy data to host")){\
		RETMACRO;}								\
}while(0)
#define CUFREE(var) cudaFree(var)

//   arr
__global__ void normalize_vec(float *arr, int arrsize, int blksize, float *maxval){
	__shared__ float max[BLKSZ];
	int idx = threadIdx.x;
	int b_beg = idx * blksize;
	if(b_beg >= arrsize) return;
	int b_end = b_beg + blksize;
	if(b_end > arrsize) b_end = arrsize;
	int i; float *ptr = &arr[b_beg];
	float *maxptr = &max[idx];
	*maxptr = *ptr++;
	for(i = b_beg +1 ; i < b_end; i++, ptr++)
		if(*maxptr < *ptr) *maxptr = *ptr;
	__syncthreads();
	if(idx == 0)
		for(i = 1; i < BLKSZ; i++)
			if(max[0] < max[i]) max[0] = max[i];
	__syncthreads();
	ptr = &arr[b_beg];
	*maxval = max[0];
	for(i = b_beg ; i < b_end; i++, ptr++)
		*ptr /= max[0];
}

//    sincos
__global__ void fill_sincos(int angles,
							float *Sin_d, float *Cos_d,
							float anglestep, float conv){
	int k = blockIdx.x * blockDim.x + threadIdx.x;
	if(k >= angles) return;
	float theta = ((float)k / anglestep - 90.f) * conv;
	sincosf(theta, &Sin_d[k], &Cos_d[k]);
}

//      
extern "C" void init_sincos(int angles){
	float anglestep = (float)angles / 180.f; //   
	float conv = M_PI / 180.f;
	int blks = (angles + BLKSZ - 1) / BLKSZ;
	int threads = BLKSZ;
	if(!Sin_d || !Cos_d || angles != sincosize){ //     
		CUFREE(Cos_d);
		CUFREE(Sin_d);
		printf("aloca %d\n", angles);
		CUALLOC(Sin_d, angles*sizeof(float));
		CUALLOC(Cos_d, angles*sizeof(float));
		fill_sincos<<<blks, threads>>>(angles, Sin_d, Cos_d, anglestep, conv);
		sincosize = angles;
	}
	cudaThreadSynchronize();
}

/*
 *      
 */
__global__ void fill_lin_hough_array(float *ima_d,
						int imW, int imH,
						float *Sin_d, float *Cos_d,
						int Rmax, int angles,
						float *hough_d){
	int xi = blockIdx.x * blockDim.x + threadIdx.x;
	int yi = blockIdx.y * blockDim.y + threadIdx.y;
	int i = xi + imW * yi;
	float x = (float)xi;
	float y = (float)yi;
	int Y = 0, k, R;
	if(xi >= imW || yi >= imH) return;
	float ima = ima_d[i];
	if(ima > 0.1f){
		for(k = 0; k < angles; k++, Y+=Rmax){
			// R = x*cos(theta) + y*sin(theta)
			R = (int)(0.5f + x * Cos_d[k] + y * Sin_d[k]);
			//if(R > 0 && R < Rmax) atomicAdd(&hough_d[R + Y], ima);
			if(R > 0 && R < Rmax) hough_d[R + Y] += ima;
		}
	}
}
/*
 *      
 * :
 * 		ima -  
 * 		imW, imH -    
 * 		Rmax -   R
 * 		angles -    (    180/angles )
 * : hough -    ,
 * 		     
 * 		!!!         
 *      
 */
extern "C" int fill_hough_lines(float *ima, int imW, int imH, int Rmax, int angles, float *hough){
	#undef RETMACRO
	#define RETMACRO do{ ret = 0; goto free_all; }while(0)
	//int i;
	int sz, ret = 1;
	int lblksz = BLKSZ;
	float  *ima_d = NULL, *hough_d = NULL, *max_d = NULL, max;
	float t0, t1;
	init_sincos(angles);
	sz = imW * imH;
	dim3 blkdim(BLKSZ, BLKSZ);
	dim3 griddim((imW+BLKSZ-1)/BLKSZ, (imH+BLKSZ-1)/BLKSZ);
CUALLOC(max_d, sizeof(float));
	CUALLOC(ima_d, sz*sizeof(float));
	//cudaHostGetDevicePointer(&ima_d, ima, 0);
	//cudaHostGetDevicePointer(&hough_d, hough, 0);
	CUMOV2DEV(ima_d, ima, sz*sizeof(float));
	sz = Rmax * angles;
	CUALLOC(hough_d, sz*sizeof(float));
	t0 = ftime();
	fill_lin_hough_array<<<griddim, blkdim>>>(ima_d, imW, imH, Sin_d,
									Cos_d, Rmax, angles, hough_d);
	cudaThreadSynchronize();
	t1 = ftime();
	printf("\nHOUGH time = %gus\n" , t1 - t0);
	t0 = t1;
	normalize_vec<<<1, lblksz>>>(hough_d, sz, (sz + lblksz - 1)/ lblksz, max_d);
	cudaThreadSynchronize();
	t1 = ftime();
	CUMOV2HOST(&max, max_d, sizeof(float));
	printf("NORM time = %gus\nmax=%g\n" , t1 - t0, max);
	CUMOV2HOST(hough, hough_d, sz*sizeof(float));
free_all:
	CUFREE(hough_d);
	CUFREE(ima_d);
CUFREE(max_d);
	return ret;
	#undef RETMACRO
	#define RETMACRO return
}
