//      imtools.c -   
//
//      Copyright 2011 Edward V. Emelianoff <eddy@sao.ru>
//
//      This program is free software; you can redistribute it and/or modify
//      it under the terms of the GNU General Public License as published by
//      the Free Software Foundation; either version 2 of the License, or
//      (at your option) any later version.
//
//      This program is distributed in the hope that it will be useful,
//      but WITHOUT ANY WARRANTY; without even the implied warranty of
//      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//      GNU General Public License for more details.
//
//      You should have received a copy of the GNU General Public License
//      along with this program; if not, write to the Free Software
//      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
//      MA 02110-1301, USA.
#include <fitsio.h>
#include "fitsheaders.h"
#include "imtools.h"
#include "opengl.h"
#include "gauss.h"
#include "CUtools.h"

//  , 256   
const int HIST_SIZE	= 256;
const int MIN_SPOT_NO = 250;	//   
const int MAX_SPOT_NO = 258;	//   

Spots *spots = NULL;

void destroy_image(IMAGE *image){
	if(!image) return;
	free(image->data);
	g_free(image->filename);
	if(image->store)
		gtk_list_store_clear(image->store);
	free(image);
}

#ifdef LEPTONICA_FOUND
/*
 *   ima (   treshold_image)
 *     pixConnComp  
 */
PIX *conv_image2pix(IMAGE *ima){
	PIX *pixs;
	l_uint32 byte, *pptr;
	GLfloat *iptr, *idata;
	int w, h, wpl, i, j, k, N;
	w = ima->width;
	h = ima->height;
	pixs = pixCreate(w, h, 1);
	wpl = pixGetWpl(pixs);
	pptr = pixGetData(pixs);
	idata = ima->data;
	for(i = h - 1; i > -1; i--){ //   
		N = 0;
		iptr = &idata[i * w];
		for(j = 0; j < wpl; j++){ //   32 
			byte = 0;
			for(k = 0; k < 32; k++){ //    32- 
				byte <<= 1;
				//         50%
				if(N++ < w && *iptr++ > 0.5)
					byte |= 1; //   
			}
			*pptr++ = byte;
		}
	}
	return pixs;
}

gboolean check_box(Window *window, BOX *box, int *X0, int *Y0, int *X, int *Y){
	if(!window->image->data) return FALSE;
	int WD = window->image->width, HT = window->image->height - 1;
	if(!(box && X0 && Y0 && X && Y)) return FALSE;
	*X0 = box->x;
	*Y0 = box->y;
	*X  = *X0 + box->w;
	*Y  = *Y0 + box->h; //   
	if(box->w < 5 || box->h < 5) return FALSE;
	if(*X0 < 0 || *X0 > WD || *X < 0 || *X > WD ||
	   *Y0 < 0 || *Y0 > HT || *Y < 0 || *Y > HT)
			return FALSE;
	return TRUE;
}
/*
 *    box   ima
 *     , 
 *     ( X)   ( Y)
 */
gboolean get_gaussian_center(Window *window, BOX *box, Coordinates *crds){
	int WD = window->image->width, HT = window->image->height - 1;
	int x, y;
	GLfloat *data;
	Points row, col;
	Point *rptr, *cptr;
	double x0, sigma;
	int X0, Y0, X, Y;
	#ifndef GSL_FOUND
		return FALSE;
	#endif
	if(!check_box(window, box, &X0, &Y0, &X, &Y)) return FALSE;
	if(!crds) return FALSE;
	row.data = calloc(box->w, sizeof(Point));
	row.n = box->w;
	col.data = calloc(box->h, sizeof(Point));
	col.n = box->h;
	cptr = col.data;
	for(y = Y0; y < Y; y++, cptr++){
		data = &window->image->data[(HT - y) * WD + X0];
		cptr->x  = (double)y;
		rptr = row.data;
		for(x = X0; x < X; x++, data++, rptr++){
			double d = (double) *data;
			rptr->y += d;
			cptr->y += d;
		}
	}
	rptr = row.data;
	for(x = X0; x < X; x++, rptr++) rptr->x  = (double)x;
	//        
	gauss_fit(&row, NULL, NULL, &sigma, &x0);
	crds->x = x0; crds->rx = sigma;
	gauss_fit(&col, NULL, NULL, &sigma, &x0);
	crds->y = x0; crds->ry = sigma;
	free(row.data); free(col.data);
	return TRUE;
}

//  -    i
// http://graphics.stanford.edu/~seander/bithacks.html#CountBitsSetParallel
// http://en.wikipedia.org/wiki/Hamming_weight
int count_bits(l_uint32 i){
	i = i - ((i >> 1) & 0x55555555);
	i = (i & 0x33333333) + ((i >> 2) & 0x33333333);
	return ((i + ((i >> 4) & 0xF0F0F0F)) * 0x1010101) >> 24;
}

gboolean get_circle_center(Window *window, BOX *box, PIX *pix, Coordinates *crds){
	int WD = window->image->width, HT = window->image->height - 1;
	int N = 0, X0, Y0, X, Y, n;
	l_uint32 *pixdata, *pixptr;
	l_int32 w, h, d, wpl, sz, i, x, y;
	double Xcenter, Ycenter, Radius, xx, yy, Intens, X2, Y2;
	GLfloat *data;
	if(!check_box(window, box, &X0, &Y0, &X, &Y)) return FALSE;
	if(!(pix && crds)) return FALSE;
	pixGetDimensions(pix, &w, &h, &d);
	if(d != 1) return FALSE;
	wpl = pixGetWpl(pix);
	pixdata = pixGetData(pix);
	pixptr = pixdata;
	sz = wpl * h;
	for(i = 0; i < sz; i++) N += count_bits(*pixptr++);
	if(N < 200000) return FALSE;
	n = 0;
	Xcenter = (double)(X0 + X) / 2.;
	Ycenter = (double)(Y0 + Y) / 2.;
	Radius = (double)(w + h) / 4.;
	DBG("x0=%g, y0=%g, r0=%g, N=%d", Xcenter, Ycenter, Radius, N);
	Xcenter = Ycenter = X2 = Y2 = Intens = 0.;
	AxisX = 0.; AxisY = (double)HT;
	for(y = 0; y < h; y++){
		yy = (double)(HT - y - Y0);
		data = &window->image->data[(HT - y - Y0) * WD + X0];
		l_uint32 *line = pixdata + y * wpl;
		for(x = 0; x < w; x++, data++){
			if(GET_DATA_BIT(line, x)){
				double d = *data;
				xx= (X0 + x);
				n++;
				Intens += d;
				Xcenter += xx * d;
				Ycenter += yy * d;
				X2 += xx*xx*d;
				Y2 += yy*yy*d;
			}
		}
	}
	xx = Xcenter / Intens;
	yy = Ycenter / Intens;
	crds->x = xx;
	crds->y = yy;
	crds->rx= sqrt(X2 / Intens - xx*xx);
	crds->ry= sqrt(Y2 / Intens - yy*yy);
	DBG("xx: %g, yy: %g, rx:%g, ry: %g", xx, yy, crds->rx, crds->ry);
	return TRUE;
}

void get_spots_coords(Window *window, IMAGE *ima){
	BOXA *boxa;
	BOX *box;
	PIX *pixs;
	Spot spot;
	Coordinates scrds;
	int n, i;
	float xc, yc;
	if(!( (window->context->transformMask & TRANS_TRES) && window->image_transformed->data)){
		g_err(_("Define treshold limits first"));
		window->context->graphMode = GR_HISTOGRAM;
		show_histogram(window);
		return;
	}
	pixs = conv_image2pix(ima);

	boxa = pixConnComp(pixs, NULL, 8);

	n = boxaGetCount(boxa);

	DBG(RED "\n\nNumber of boxes: %d\n" OLDCOLOR, n);
	if(n > MAX_SPOT_NO){
		g_err(_("Too many points"));
		return;
	}
	if(n < MIN_SPOT_NO){
		g_err(_("Not enough points"));
		return;
	}
	spots_free(&spots);
	if(!(spots = spots_alloc(n))) return;
	#ifndef GSL_FOUND
		g_err(_("No GSL library found, using simple centroids method"));
	#endif
	for(i = 0; i < n; i++){
		if((box = boxaGetBox(boxa, i, L_CLONE)) == NULL){
			g_err(_("Box not found"));
			continue;
		}
		if(!get_gaussian_center(window, box, &scrds)){
			scrds.rx = ((double)box->w-1.)/2.;
			scrds.ry = ((double)box->h-1.)/2.;
			xc = box->x + scrds.rx;
			yc = box->y + scrds.ry;
			scrds.x = xc; scrds.y = yc;
		}
		spot.c = scrds; spot.id = 0;
		spot.box.x = box->x;
		spot.box.y = box->y;
		spot.box.w = box->w;
		spot.box.h = box->h;
		spots_add(spots, &spot, BY_COPY);
		boxDestroy(&box);
	}
	sort_spots(spots);

	boxaDestroy(&boxa);
	gdk_window_invalidate_rect(window->drawingArea->window,
		&window->drawingArea->allocation, FALSE);
	DBG("return");
}

void get_circles_params(Window *window, IMAGE *ima){
	BOXA *boxa;
	PIXA *pixa;
	BOX *box;
	PIX *pixs, *pix;
	Spot spot;
	Coordinates scrds;
	int n, i;
	if(!( (window->context->transformMask & TRANS_TRES) && window->image_transformed->data)){
		g_err(_("Define treshold limits first"));
		window->context->graphMode = GR_HISTOGRAM;
		show_histogram(window);
		return;
	}
	pixs = conv_image2pix(ima);

	boxa = pixConnComp(pixs, &pixa, 8);
	n = boxaGetCount(boxa);

	DBG(RED "\n\nNumber of boxes: %d\n" OLDCOLOR, n);
	spots_free(&spots);
	if(!(spots = spots_alloc(n))) return;
	for(i = 0; i < n; i++){
		if((box = boxaGetBox(boxa, i, L_CLONE)) == NULL){
			g_err(_("Box not found"));
			continue;
		}
		if((pix = pixaGetPix(pixa, i, L_CLONE)) == NULL){
			g_err(_("Pix not found"));
			continue;
		}
		if(get_circle_center(window, box, pix, &scrds)){
			spot.c = scrds; spot.id = i;
			spot.box.x = box->x;
			spot.box.y = box->y;
			spot.box.w = box->w;
			spot.box.h = box->h;
			spots_add(spots, &spot, BY_COPY);
		}
		boxDestroy(&box);
	}

	boxaDestroy(&boxa);
	pixaDestroy(&pixa);

	gdk_window_invalidate_rect(window->drawingArea->window,
		&window->drawingArea->allocation, FALSE);
	DBG("return");
}
#else // LEPTONICA_FOUND
void get_spots_coords(Window *window, IMAGE *ima){
	LEPTERR;
	if(window || ima) return;
}
gboolean get_gaussian_center(Window *window, BOX *box, Coordinates *crds){
	LEPTERR;
	if(window || box || crds) return FALSE;
	return FALSE;
}
PIX *conv_image2pix(IMAGE *ima){
	LEPTERR;
	if(ima) return NULL;
	return NULL;
}
void get_circles_params(Window *window, IMAGE *ima){
	LEPTERR;
	if(window || ima) return;
}
#endif  // LEPTONICA_FOUND


/*
 *    ima
 *    256  
 */
gboolean make_histogram(IMAGE *ima){
	FNAME();
	ssize_t i, sz;
	int j, pt, hmax, *hdata, max, min;
	GLfloat *ptr, hsize;
	if(!ima->data){
		g_err(_("Image is empty"));
		return FALSE;
	}
	free(ima->stat.histogram.data);
	ima->stat.histogram.data = NULL;
	//      HIST_SIZE   
	ima->stat.histogram.data = calloc(HIST_SIZE,  sizeof(int));
	if(!ima->stat.histogram.data){
		g_err(_("Can't allocate memory for histogram"));
		return FALSE;
	}
	hdata = ima->stat.histogram.data;
	ima->stat.histogram.size = HIST_SIZE;
	hsize = (GLfloat) HIST_SIZE;
	hmax = HIST_SIZE - 1;
	ptr = ima->data;
	sz = (ssize_t)(ima->width) * (ssize_t)(ima->height);
	for(i = 0; i < sz; i++, ptr++){
		pt = (int)(*ptr * hsize);
		if(pt < 0) pt = 0;
		else if(pt > hmax) pt = hmax;
		hdata[pt]++;
	}
	max = min = hdata[0];
	for(j = 1; j < HIST_SIZE; j++){
		pt = hdata[j];
		if(max < pt) max = pt;
		else if(min > pt) min = pt;
	}
	ima->stat.histogram.max = max;
	ima->stat.histogram.min = min;
	ima->stat.histogram.bot_tres = -1.;
	ima->stat.histogram.top_tres = -1.;
	return TRUE;
}

/*
 *   image_transformed    
 * bottom -  , top -  
 * ,  bottom,   
 * ,  top,  
 *  -   ( )
 */
void treshold_image(Window *window, GLfloat bottom, GLfloat top){
	FNAME();
	ssize_t i, sz;
	IMAGE *ima = window->image;
	IMAGE *i_new = window->image_transformed;
	GLfloat *p_old, *p_new, pt, wd;
	if(!(p_old = ima->data)){
		g_err(_("Image is empty"));
		return;
	}
	free(i_new->data);
	if((top - bottom) < 0.01) top = bottom + 0.01;
	sz = (ssize_t)(ima->width) * (ssize_t)(ima->height);
	i_new->data = malloc(sz * sizeof(GLfloat));
	i_new->width = ima->width; i_new->height = ima->height;
	i_new->bZero = 0.; i_new->bScale = 1.;
	i_new->maxVal = ima->maxVal;
	i_new->stat.max = top * ima->stat.max;
	i_new->stat.min = 0.;
	p_new = i_new->data;
	wd = top - bottom;
	for(i = 0; i < sz; i++, p_new++, p_old++){
		pt = *p_old;
		if(pt < bottom) pt = 0.;
		else if(pt > top) pt = 1.;
		else pt = (pt - bottom) / wd;
		*p_new = pt;
	}
	window->context->transformMask |= TRANS_TRES;
	gen_texture(window->image_transformed, window, TRUE);
}

const int SINCOSSIZE = 540; //     
							// ( [-90  +180) .)

void hough_lines(Window *window){ //   'i'
	FNAME();
	int w, h, Rmax, sz;
	double hd, wd;
	IMAGE *ima = window->image;
	Window *outwin = init_window(window, OPENGL_WINDOW);
	IMAGE *i_new = outwin->image;
	if(!ima->data){
		g_err(_("Image is empty"));
		return;
	}
	free(i_new->data);
	w = ima->width; h = ima->height;
	wd = (double)w; hd = (double)h;
	Rmax = (int)(0.5 + sqrt(wd*wd + hd*hd));
	sz = SINCOSSIZE * Rmax;
	DBG("sz: %d", sz);
	i_new->data = calloc(sz, sizeof(GLfloat));
	if(!i_new->data){
		g_err(_("Can't allocate memory for Hough transform"));
		return;
	}
	i_new->width = Rmax; i_new->height = SINCOSSIZE;
	i_new->bZero = 0.; i_new->bScale = 1.;
	i_new->maxVal = 1.; i_new->stat.min = 0.;
	i_new->stat.max = -1.;
	i_new->dtype = FLOAT_IMG;
	gchar *str = malloc(256);
	snprintf(str, 255, "Hough transform of file %s", ima->filename);
	init_keylist(i_new);
	add_key(i_new, "TSTRING", "OBJECT", "Hough", str, TSTRING);
	free(str);
#ifdef EBUG
	double t0 = dtime();
#endif
	if(!fill_hough_lines(ima->data, w, h, Rmax, SINCOSSIZE, i_new->data)){
		g_err(_("Error in Hough transform module"));
		destroy_window(outwin);
		return;
	}
	DBG("Hough transform for lines, time: %gs", dtime() - t0);
	//Context->transformMask |= TRANS_TRES;
	//DBG("HOUGH: maxR=%d, size=%d, max: %g at R=%d, theta=%g",
		//i_new->width, i_new->height, max, R%Rmax, ((double)(R/Rmax))/2.-90.);
	gen_texture(outwin->image, outwin, FALSE);
	force_redraw(outwin->drawingArea);
	outwin->image->imagetype = HOUGH;
	CH_WIN_TITLE(outwin, "%s (Hough) %s", PRGNAME, window->image->filename);
}

/*
 *   (x,y)       
 *    (R,phi)     
 *   
 */
void get_houg_line( Window *window,
					double x, double y, //   
					double *phi_, double *R_ //   
					){
	double R, phi, ang_sep, sinphi, cosphi;
	double x1=0.,y1=0., x2=0.,y2=0., yl,yr, xu,xd, X1, X2, Y1, Y2;
	if(!window || !window->parent) return;
	double W = window->parent->image->width-1., H = window->parent->image->height-1.;
	int rdy = 0;
	IMAGE *image = window->image;
	conv_mouse_to_image_coords(x, y, &R, &phi, window);
	ang_sep = 270. / (double)image->height / 180. * M_PI;
	phi = phi*ang_sep - M_PI/2.;
	sinphi = sin(phi); cosphi = cos(phi);
	phi *= 180./M_PI; //   
	DBG("line: phi = %g, R = %g", phi, R);
	*phi_ = phi; *R_ = R;
	GdkDrawable *d = window->parent->drawingArea->window;
	static GdkGC *gc = NULL;
	if(!gc){
		gc = gdk_gc_new(d);
		gdk_gc_set_foreground(gc, &(window->parent->drawingArea->style->white));
		gdk_gc_set_function(gc, GDK_XOR);
	}
	yl = R/sinphi; yr = (R-W*cosphi)/sinphi; //     
	xd = R/cosphi; xu = (R-H*sinphi)/cosphi; //    
	DBG("yl=%g, yr=%g, xd=%g, xu=%g",yl,yr,xd,xu);
	if(yl < H && yl > -0.5){
		DBG("left selected");
		y1 = yl; x1 = 0.; rdy = 1;
	}
	if(yr < H && yr > -0.5){
		DBG("right selected");
		if(rdy == 0){
			y1 = yr; x1 = W; rdy = 1;
		}else{
			y2 = yr; x2 = W; rdy = 2; goto lines_selected;
		}
	}
	if(xd < W && xd > -0.5){
		DBG("down selected");
		if(rdy == 0){
			x1 = xd; y1 = 0.; rdy = 1;
		}else{
			x2 = xd; y2 = 0.; rdy = 2; goto lines_selected;
		}
	}
	if(xu < W && xu > -0.5){
		DBG("up selected");
		if(rdy == 0){
			x1 = xu; y1 = H; rdy = 1;
		}else{
			x2 = xu; y2 = H; rdy = 2;
		}
	}
lines_selected:
	if(rdy != 2) DBG("WTF? not two points");
	else{
		DBG("line in image CS: (%g, %g) - (%g, %g)", x1,y1,x2,y2);
		conv_image_to_mouse_coords(x1, y1, &X1, &Y1, window->parent);
		conv_image_to_mouse_coords(x2, y2, &X2, &Y2, window->parent);
		gdk_draw_line(d, gc, X1,Y1, X2,Y2);
		DBG("line in window CS: (%g, %g) - (%g, %g)", X1,Y1,X2,Y2);
	}
}
